"""
Author: Keith Bourgoin, Emmett Butler
"""
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
__all__ = ["Topic"]
import datetime as dt
import logging
from collections import defaultdict

from .balancedconsumer import BalancedConsumer
from .common import OffsetType, EPOCH
from .exceptions import LeaderNotFoundError
from .managedbalancedconsumer import ManagedBalancedConsumer
from .partition import Partition
from .producer import Producer
from .protocol import PartitionOffsetRequest
from .simpleconsumer import SimpleConsumer
from .utils.compat import iteritems, itervalues
try:
    from .handlers import GEventHandler
except ImportError:
    GEventHandler = None


log = logging.getLogger(__name__)


try:
    from . import rdkafka
    log.info("Successfully loaded pykafka.rdkafka extension.")
except ImportError:
    rdkafka = False
    log.info("Could not load pykafka.rdkafka extension.")
    log.debug("Traceback:", exc_info=True)


class Topic(object):
    """
    A Topic is an abstraction over the kafka concept of a topic.
    It contains a dictionary of partitions that comprise it.
    """
    def __init__(self, cluster, topic_metadata):
        """Create the Topic from metadata.

        :param cluster: The Cluster to use
        :type cluster: :class:`pykafka.cluster.Cluster`
        :param topic_metadata: Metadata for all topics.
        :type topic_metadata: :class:`pykafka.protocol.TopicMetadata`
        """
        self._name = topic_metadata.name
        self._cluster = cluster
        self._partitions = {}
        self.update(topic_metadata)

    def __repr__(self):
        return "<{module}.{classname} at {id_} (name={name})>".format(
            module=self.__class__.__module__,
            classname=self.__class__.__name__,
            id_=hex(id(self)),
            name=self._name
        )

    @property
    def name(self):
        """The name of this topic"""
        return self._name

    @property
    def partitions(self):
        """A dictionary containing all known partitions for this topic"""
        return self._partitions

    def get_producer(self, use_rdkafka=False, **kwargs):
        """Create a :class:`pykafka.producer.Producer` for this topic.

        For a description of all available `kwargs`, see the Producer docstring.
        """
        if not rdkafka and use_rdkafka:
            raise ImportError("use_rdkafka requires rdkafka to be installed")
        if GEventHandler and isinstance(self._cluster.handler, GEventHandler) and use_rdkafka:
            raise ImportError("use_rdkafka cannot be used with gevent")
        Cls = Producer
        if rdkafka and use_rdkafka:
            Cls = rdkafka.RdKafkaProducer
            kwargs.pop('block_on_queue_full', None)
        return Cls(self._cluster, self, **kwargs)

    def get_sync_producer(self, **kwargs):
        """Create a :class:`pykafka.producer.Producer` for this topic.

        The created `Producer` instance will have `sync=True`.

        For a description of all available `kwargs`, see the Producer docstring.
        """
        return Producer(self._cluster, self, sync=True, **kwargs)

    def fetch_offset_limits(self, offsets_before, max_offsets=1):
        """Get information about the offsets of log segments for this topic

        The ListOffsets API, which this function relies on, primarily deals
        with topics in terms of their log segments. Its behavior can be summed
        up as follows: it returns some subset of starting message offsets for
        the log segments of each partition. The particular subset depends on
        this function's two arguments, filtering by timestamp and in certain
        cases, count. The documentation for this API is notoriously imprecise,
        so here's a little example to illustrate how it works.

        Take a topic with three partitions 0,1,2. 2665 messages have been
        produced to this topic, and the brokers' `log.segment.bytes` settings
        are configured such that each log segment contains roughly 530
        messages. The two oldest log segments have been deleted due to log retention
        settings such as `log.retention.hours`. Thus, the `log.dirs` currently
        contains these files for partition 0:

        /var/local/kafka/data/test2-0/00000000000000001059.log
        /var/local/kafka/data/test2-0/00000000000000002119.log
        /var/local/kafka/data/test2-0/00000000000000001589.log
        /var/local/kafka/data/test2-0/00000000000000002649.log

        The numbers on these filenames indicate the offset of the earliest
        message contained within. The most recent message was written at 1523572215.69.

        Given this log state, a call to this function with
        offsets_before=OffsetType.LATEST
        and max_offsets=100 will result in a return value of
        [2665,2649,2119,1589,1059] for partition 0. The first value (2665) is
        the offset of the latest available message from the latest log
        segment. The other four offsets are those of the earliest messages
        from each log segment for the partition. Changing max_offsets to
        3 will result in only the first three elements of this list being returned.

        A call to this function with offsets_before=OffsetType.EARLIEST will
        result in a value of [1059] - only the offset of the earliest message
        present in log segments for partition 0. In this case, the return
        value is not affected by max_offsets.

        A call to this function with offsets_before=(1523572215.69 * 1000)
        (the timestamp in milliseconds of the very last message written to the
        partition) will result in a value of [2649,2119,1589,1059]. This is
        the same list as with OffsetType.LATEST, but with the first element
        removed. This is because unlike the other elements, the message with
        this offset (2665) was not written *before* the given timestamp.

        In cases where there are no log segments fitting the given criteria
        for a partition, an empty list is returned. This applies if the given
        timestamp is before the write time of the oldest message in the
        partition, as well as if there are no log segments for the partition.

        Thanks to Andras Beni from the Kafka users mailing list for providing
        this example.

        :param offsets_before: Epoch timestamp in milliseconds or datetime indicating the
            latest write time for returned offsets. Only offsets of messages
            written before this timestamp will be returned. Permissible
            special values are `common.OffsetType.LATEST`, indicating that
            offsets from all available log segments should be returned, and
            `common.OffsetType.EARLIEST`, indicating that only the offset of
            the earliest available message should be returned. Deprecated::2.7,3.6:
            do not use int
        :type offsets_before: `datetime.datetime` or int
        :param max_offsets: The maximum number of offsets to return when more
            than one is available. In the case where `offsets_before ==
            OffsetType.EARLIEST`, this parameter is meaningless since there is
            always either one or zero earliest offsets. In other cases, this
            parameter slices off the earliest end of the list, leaving the
            latest `max_offsets` offsets.
        :type max_offsets: int
        """
        if isinstance(offsets_before, dt.datetime):
            offsets_before = round((offsets_before - EPOCH).total_seconds() * 1000)
        requests = defaultdict(list)  # one request for each broker
        for part in itervalues(self.partitions):
            requests[part.leader].append(PartitionOffsetRequest(
                self.name, part.id, offsets_before, max_offsets
            ))
        output = {}
        for broker, reqs in iteritems(requests):
            res = broker.request_offset_limits(reqs)
            output.update(res.topics[self.name])
        return output

    def earliest_available_offsets(self):
        """Get the earliest offset for each partition of this topic."""
        return self.fetch_offset_limits(OffsetType.EARLIEST)

    def latest_available_offsets(self):
        """Fetch the next available offset

        Get the offset of the next message that would be appended to each partition of
            this topic.
        """
        return self.fetch_offset_limits(OffsetType.LATEST)

    def update(self, metadata):
        """Update the Partitions with metadata about the cluster.

        :param metadata: Metadata for all topics
        :type metadata: :class:`pykafka.protocol.TopicMetadata`
        """
        p_metas = metadata.partitions

        # Remove old partitions
        removed = set(self._partitions.keys()) - set(p_metas.keys())
        if len(removed) > 0:
            log.info('Removing %d partitions', len(removed))
        for id_ in removed:
            log.debug('Removing partition %s', self._partitions[id_])
            self._partitions.pop(id_)

        # Add/update current partitions
        brokers = self._cluster.brokers
        if len(p_metas) > 0:
            log.info("Adding %d partitions", len(p_metas))
        for id_, meta in iteritems(p_metas):
            if meta.leader not in brokers:
                raise LeaderNotFoundError()
            if meta.id not in self._partitions:
                log.debug('Adding partition %s/%s', self.name, meta.id)
                self._partitions[meta.id] = Partition(
                    self, meta.id,
                    brokers[meta.leader],
                    # only add replicas that the cluster is aware of to avoid
                    # KeyErrors here. inconsistencies will be automatically
                    # resolved when `Cluster.update` is called.
                    [brokers[b] for b in meta.replicas if b in brokers],
                    [brokers[b] for b in meta.isr],
                )
            else:
                self._partitions[id_].update(brokers, meta)

    def get_simple_consumer(self,
                            consumer_group=None,
                            use_rdkafka=False,
                            **kwargs):
        """Return a SimpleConsumer of this topic

        :param consumer_group: The name of the consumer group to join
        :type consumer_group: bytes
        :param use_rdkafka: Use librdkafka-backed consumer if available
        :type use_rdkafka: bool
        """
        if not rdkafka and use_rdkafka:
            raise ImportError("use_rdkafka requires rdkafka to be installed")
        if GEventHandler and isinstance(self._cluster.handler, GEventHandler) and use_rdkafka:
            raise ImportError("use_rdkafka cannot be used with gevent")
        Cls = (rdkafka.RdKafkaSimpleConsumer
               if rdkafka and use_rdkafka else SimpleConsumer)
        return Cls(self,
                   self._cluster,
                   consumer_group=consumer_group,
                   **kwargs)

    def get_balanced_consumer(self, consumer_group, managed=False, **kwargs):
        """Return a BalancedConsumer of this topic

        :param consumer_group: The name of the consumer group to join
        :type consumer_group: bytes
        :param managed: If True, manage the consumer group with Kafka using the 0.9
            group management api (requires Kafka >=0.9))
        :type managed: bool
        """
        if not managed:
            if "zookeeper_connect" not in kwargs and \
                    self._cluster._zookeeper_connect is not None:
                kwargs['zookeeper_connect'] = self._cluster._zookeeper_connect
            cls = BalancedConsumer
        else:
            cls = ManagedBalancedConsumer
        return cls(self, self._cluster, consumer_group, **kwargs)



import time
import os

from pykafka.test.kafka_instance import KafkaInstance, KafkaConnection


def get_cluster():
    """Gets a Kafka cluster for testing, using one already running is possible.

    An already-running cluster is determined by environment variables:
    BROKERS, ZOOKEEPER, KAFKA_BIN.  This is used primarily to speed up tests
    in our Travis-CI environment.
    """
    if os.environ.get('BROKERS', None) and \
       os.environ.get('ZOOKEEPER', None) and \
       os.environ.get('KAFKA_BIN', None):
        # Broker is already running. Use that.
        return KafkaConnection(os.environ['KAFKA_BIN'],
                               os.environ['BROKERS'],
                               os.environ['ZOOKEEPER'],
                               os.environ.get('BROKERS_SSL', None))
    else:
        return KafkaInstance(num_instances=3)


def stop_cluster(cluster):
    """Stop a created cluster, or merely flush a pre-existing one."""
    if isinstance(cluster, KafkaInstance):
        cluster.terminate()
    else:
        cluster.flush()


def retry(assertion_callable, retry_time=10, wait_between_tries=0.1, exception_to_retry=AssertionError):
    """Retry assertion callable in a loop"""
    start = time.time()
    while True:
        try:
            return assertion_callable()
        except exception_to_retry as e:
            if time.time() - start >= retry_time:
                raise e
            time.sleep(wait_between_tries)



import logging
from pkg_resources import parse_version

from pykafka.exceptions import RdKafkaStoppedException, ProducerStoppedException
from pykafka.producer import Producer, CompressionType
from pykafka.utils.compat import get_bytes
from . import _rd_kafka
from . import helpers


log = logging.getLogger(__name__)


class RdKafkaProducer(Producer):
    """A librdkafka-backed version of pykafka.Producer

    This aims to conform to the Producer interface as closely as possible.
    For an overview of how configuration keys are mapped to librdkafka's, see
    _mk_rdkafka_config_lists.

    Note: the `retry_backoff_ms` parameter behaves slightly differently in the
    `RdKafkaProducer` than it does in the pure Python `Producer`. In the rdkafka
    implementation, `retry_backoff_ms` indicates the exact time spent between message
    resend attempts, but in the pure Python version the time between attempts is also
    influenced by several other parameters, including `linger_ms` and `socket_timeout_ms`.

    Note also that `linger_ms` becomes the rdkafka-based producer's
    `queue.buffering.max.ms` config option and thus must be within the acceptable range
    defined by librdkafka. Several other parameters, including `ack_timeout_ms`,
    `max_retries`, and `retry_backoff_ms`, are used to derive certain librdkafka
    config values. Certain combinations of values for these parameters can result in
    configuration errors from librdkafka.

    The `broker_version` argument on `KafkaClient` must be set correctly to use the
    rdkafka producer.
    """
    def __init__(self,
                 cluster,
                 topic,
                 partitioner=None,
                 compression=CompressionType.NONE,
                 max_retries=3,
                 retry_backoff_ms=100,
                 required_acks=1,
                 ack_timeout_ms=10 * 1000,
                 max_queued_messages=100000,
                 min_queued_messages=2000,  # NB differs from pykafka.Producer
                 linger_ms=5 * 1000,
                 max_request_size=1000012,
                 sync=False,
                 delivery_reports=False,
                 auto_start=True,
                 serializer=None):
        callargs = {k: v for k, v in vars().items()
                    if k not in ("self", "__class__")}
        self._broker_version = cluster._broker_version
        self._rdk_producer = None
        self._poller_thread = None
        self._stop_poller_thread = cluster.handler.Event()
        # super() must come last because it calls start()
        super(RdKafkaProducer, self).__init__(**callargs)

    def start(self):
        if not self._running:
            brokers = b','.join(b.host + b":" + get_bytes(str(b.port))
                                for b in self._cluster.brokers.values())
            conf, topic_conf = self._mk_rdkafka_config_lists()

            self._rdk_producer = _rd_kafka.Producer()
            self._rdk_producer.configure(conf=conf)
            self._rdk_producer.configure(topic_conf=topic_conf)
            self._rdk_producer.start(
                brokers, self._topic.name, self._delivery_reports.put)
            self._running = True

            def poll(rdk_handle, stop_event):
                while not stop_event.is_set() or rdk_handle.outq_len() > 0:
                    rdk_handle.poll(timeout_ms=1000)
                log.debug("Exiting RdKafkaProducer poller thread cleanly.")

            self._stop_poller_thread.clear()
            self._poller_thread = self._cluster.handler.spawn(
                poll, args=(self._rdk_producer, self._stop_poller_thread))

    def stop(self):
        self._stop_poller_thread.set()
        super(RdKafkaProducer, self).stop()
        self._rdk_producer.stop()

    def _produce(self, message):
        try:
            self._rdk_producer.produce(message)
        except RdKafkaStoppedException:
            raise ProducerStoppedException

    def _wait_all(self):
        log.info("Blocking until all messages are sent")
        if self._poller_thread is not None:
            self._poller_thread.join()

    def _mk_rdkafka_config_lists(self):
        """Populate conf, topic_conf to configure the rdkafka producer"""
        map_compression_types = {
            CompressionType.NONE: "none",
            CompressionType.GZIP: "gzip",
            CompressionType.SNAPPY: "snappy",
            CompressionType.LZ4: "lz4",
        }

        # For documentation purposes, all producer-relevant settings (all those
        # marked 'P' or '*') that appear in librdkafka/CONFIGURATION.md should
        # be listed below, in either `conf` or `topic_conf`, even if we do not
        # set them and they are commented out.
        ver10 = parse_version(self._broker_version) >= parse_version("0.10.0")
        conf = {  # destination: rd_kafka_conf_set
            "client.id": "pykafka.rdkafka",
            # Handled via rd_kafka_brokers_add instead:
            # "metadata.broker.list"

            # NB these refer not to payloads, but to wire messages
            # We've no real equivalents for these, but defaults should be fine:
            # "receive.message.max.bytes"

            # No direct equivalents:
            # "metadata.request.timeout.ms"
            # "topic.metadata.refresh.interval.ms"
            # "topic.metadata.refresh.fast.cnt"
            # "topic.metadata.refresh.fast.interval.ms"
            # "topic.metadata.refresh.sparse"

            # "debug": "all",

            "socket.timeout.ms": self._cluster._socket_timeout_ms,
            # "socket.send.buffer.bytes"
            # "socket.receive.buffer.bytes"
            # "socket.keepalive.enable"
            # "socket.max.fails"
            # "broker.address.ttl"
            # "broker.address.family"

            # None of these need to be hooked up
            # "statistics.interval.ms"
            # "error_cb"  # we let errors be reported via log_cb
            # "stats_cb"

            # "log_cb"  # gets set in _rd_kafka module
            # "log_level": 7,

            # "socket_cb"
            # "open_cb"
            # "opaque"
            # "internal.termination.signal"
            # Do not log connection.close, which may be caused by
            # connections.max.idle.ms
            # Cf https://github.com/edenhill/librdkafka/issues/437
            "log.connection.close": "false",

            "queue.buffering.max.messages": self._max_queued_messages,
            "queue.buffering.max.ms": self._linger_ms,
            "message.send.max.retries": self._max_retries,
            "retry.backoff.ms": self._retry_backoff_ms,
            "compression.codec": map_compression_types[self._compression],
            "batch.num.messages": self._min_queued_messages,
            "message.max.bytes": self._max_request_size,

            # Report successful and failed messages so we know to dealloc them
            "delivery.report.only.error": "false",
            "api.version.request": ver10,
            # "dr_cb"
            # "dr_msg_cb"  # gets set in _rd_kafka module
        }
        # broker.version.fallback is incompatible with >-0.10
        if not ver10:
            conf["broker.version.fallback"] = self._broker_version
        conf.update(helpers.rdk_ssl_config(self._cluster))

        topic_conf = {
            # see https://github.com/edenhill/librdkafka/issues/208
            "request.required.acks": (
                self._required_acks if self._required_acks <= 1 else -1),
            # "enforce.isr.cnt"  # no current equivalent
            "request.timeout.ms": self._ack_timeout_ms,

            # This doesn't have an equivalent in pykafka.  It defaults to 300s
            # in librdkafka, which may be surprisingly long, so instead we
            # apply a formula to approach what you'd expect as a pykafka user:
            "message.timeout.ms": 1.2 * self._max_retries * (
                self._ack_timeout_ms + self._retry_backoff_ms),

            "produce.offset.report": self._delivery_reports.queue is not None,
            # "partitioner"  # dealt with in pykafka
            # "opaque"
        }
        # librdkafka expects all config values as strings:
        conf = [(key, str(conf[key])) for key in conf]
        topic_conf = [(key, str(topic_conf[key])) for key in topic_conf]
        return conf, topic_conf



__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from pkg_resources import parse_version


class Serializable(object):
    __slots__ = []

    def __len__(self):
        """Length of the bytes that will be sent to the Kafka server."""
        raise NotImplementedError()

    def pack_into(self, buff, offset):
        """Pack serialized bytes into buff starting at offset ``offset``"""
        raise NotImplementedError()


def serialize_utf8(value, partition_key):
    """A serializer accepting bytes or str arguments and returning utf-8 encoded bytes

    Can be used as `pykafka.producer.Producer(serializer=serialize_utf8)`
    """
    if value is not None and type(value) != bytes:
        # allow UnicodeError to be raised here if the encoding fails
        value = value.encode('utf-8')
    if partition_key is not None and type(partition_key) != bytes:
        partition_key = partition_key.encode('utf-8')
    return value, partition_key


def deserialize_utf8(value, partition_key):
    """A deserializer accepting bytes arguments and returning utf-8 strings

    Can be used as `pykafka.simpleconsumer.SimpleConsumer(deserializer=deserialize_utf8)`,
    or similarly in other consumer classes
    """
    # allow UnicodeError to be raised here if the decoding fails
    if value is not None:
        value = value.decode('utf-8')
    if partition_key is not None:
        partition_key = partition_key.decode('utf-8')
    return value, partition_key


VERSIONS_CACHE = {}


class ApiVersionAware(object):
    """Mixin class that facilitates standardized discovery of supported protocol versions
    """
    @classmethod
    def get_version_impl(cls, api_versions):
        """
        Return the class from `pykafka.protocol` implementing support for the highest
        version of the API supported by `cls` supported by both the calling Broker and
        pykafka itself.

        This method requires that `cls` implements the following attributes:
        * cls.get_versions() - a @classmethod taking no arguments aside from `cls` and
            returning a dictionary mapping integer API version numbers to the classes
            from `pykafka.protocol` implementing support for those versions of the
            request or response. For example:
                @classmethod
                def get_versions(cls):
                    return {0: FetchResponse, 1: FetchResponseV1, 2: FetchResponseV2}
            indicates that `FetchResponse` implements support for v0 of the fetch
            response, `FetchResponseV1` implements support for v1 of the fetch response,
            et cetera.
        * cls.API_KEY - a class attribute indicating the API key of the request or
            response. Note that `Response` instances require this attribute despite
            responses not explicitly containing the API key as defined by the Kafka
            protocol.

        :param api_versions: A sequence of :class:`pykafka.protocol.ApiVersionsSpec`
            objects indicating the API version compatibility of the calling Broker
        :type api_versions: Iterable of :class:`pykafka.protocol.ApiVersionsSpec`
        :type return: An object of the same parent class as `cls` (either
            :class:`pykafka.protocol.Request` or :class:`pykafka.protocol.Response`)
        """
        if not hasattr(cls, "get_versions") or not hasattr(cls, "API_KEY"):
            raise AttributeError("get_version_impl requires that {} implement both "
                                 "get_versions() and API_KEY.".format(cls))
        cached_version = VERSIONS_CACHE.get(cls)
        if cached_version:
            return cached_version
        sorted_versions = sorted(cls.get_versions().keys(), reverse=True)
        broker_max = api_versions[cls.API_KEY].max if api_versions else 0
        for version in sorted_versions:
            if version <= broker_max:
                highest_client_supported = cls.get_versions()[version]
                VERSIONS_CACHE[cls] = highest_client_supported
                return highest_client_supported


def msg_protocol_version(broker_version):
    if parse_version(broker_version) >= parse_version("0.10.0"):
        return 1
    return 0



# - coding: utf-8 -
import struct
from collections import namedtuple, defaultdict
from pkg_resources import parse_version

from .base import Request, Response
from .message import MessageSet
from ..common import CompressionType
from ..utils import struct_helpers, compression
from ..utils.compat import iteritems


_PartitionFetchRequest = namedtuple(
    'PartitionFetchRequest',
    ['topic_name', 'partition_id', 'offset', 'max_bytes']
)


class PartitionFetchRequest(_PartitionFetchRequest):
    """Fetch request for a specific topic/partition

    :ivar topic_name: Name of the topic to fetch from
    :ivar partition_id: Id of the partition to fetch from
    :ivar offset: Offset at which to start reading
    :ivar max_bytes: Max bytes to read from this partition (default: 300kb)
    """
    def __new__(cls, topic, partition, offset, max_bytes=1024 * 1024):
        return super(PartitionFetchRequest, cls).__new__(
            cls, topic, partition, offset, max_bytes)


class FetchRequest(Request):
    """A Fetch request sent to Kafka

    Specification::

        FetchRequest => ReplicaId MaxWaitTime MinBytes [TopicName [Partition FetchOffset MaxBytes]]
          ReplicaId => int32
          MaxWaitTime => int32
          MinBytes => int32
          TopicName => string
          Partition => int32
          FetchOffset => int64
          MaxBytes => int32
    """
    API_KEY = 1

    @classmethod
    def get_versions(cls):
        return {0: FetchRequest, 1: FetchRequest}

    def __init__(self, partition_requests=[], timeout=1000, min_bytes=1024,
                 api_version=0):
        """Create a new fetch request

        Kafka 0.8 uses long polling for fetch requests, which is different
        from 0.7x. Instead of polling and waiting, we can now set a timeout
        to wait and a minimum number of bytes to be collected before it
        returns. This way we can block effectively and also ensure good network
        throughput by having fewer, large transfers instead of many small ones
        every time a byte is written to the log.

        :param partition_requests: Iterable of
            :class:`kafka.pykafka..protocol.PartitionFetchRequest` for this request
        :param timeout: Max time to wait (in ms) for a response from the server
        :param min_bytes: Minimum bytes to collect before returning
        """
        self.timeout = timeout
        self.min_bytes = min_bytes
        self._reqs = defaultdict(dict)
        self.api_version = api_version
        for req in partition_requests:
            self.add_request(req)

    def add_request(self, partition_request):
        """Add a topic/partition/offset to the requests

        :param topic_name: The topic to fetch from
        :param partition_id: The partition to fetch from
        :param offset: The offset to start reading data from
        :param max_bytes: The maximum number of bytes to return in the response
        """
        pr = partition_request
        self._reqs[pr.topic_name][pr.partition_id] = (pr.offset, pr.max_bytes)

    def __len__(self):
        """Length of the serialized message, in bytes"""
        # replica + max wait + min bytes + len(topics)
        size = self.HEADER_LEN + 4 + 4 + 4 + 4
        for topic, parts in iteritems(self._reqs):
            # topic name + len(parts)
            size += 2 + len(topic) + 4
            # partition + fetch offset + max bytes => for each partition
            size += (4 + 8 + 4) * len(parts)
        return size

    def get_bytes(self):
        """Serialize the message

        :returns: Serialized message
        :rtype: :class:`bytearray`
        """
        output = bytearray(len(self))
        self._write_header(output, api_version=self.api_version)
        offset = self.HEADER_LEN
        struct.pack_into('!iiii', output, offset,
                         -1, self.timeout, self.min_bytes, len(self._reqs))
        offset += 16
        for topic_name, partitions in iteritems(self._reqs):
            fmt = '!h%dsi' % len(topic_name)
            struct.pack_into(
                fmt, output, offset, len(topic_name), topic_name,
                len(partitions)
            )
            offset += struct.calcsize(fmt)
            for partition_id, (fetch_offset, max_bytes) in iteritems(partitions):
                struct.pack_into('!iqi', output, offset,
                                 partition_id, fetch_offset, max_bytes)
                offset += 16
        return output


FetchPartitionResponse = namedtuple(
    'FetchPartitionResponse',
    ['max_offset', 'messages', 'err']
)


class FetchResponse(Response):
    """Unpack a fetch response from the server

    Specification::

        FetchResponse => [TopicName [Partition ErrorCode HighwaterMarkOffset MessageSetSize MessageSet]]
          TopicName => string
          Partition => int32
          ErrorCode => int16
          HighwaterMarkOffset => int64
          MessageSetSize => int32
    """
    API_VERSION = 0
    API_KEY = 1

    @classmethod
    def get_versions(cls):
        return {0: FetchResponse, 1: FetchResponseV1, 2: FetchResponseV2}

    def __init__(self, buff, offset=0, broker_version='0.9.0'):
        """Deserialize into a new Response

        :param buff: Serialized message
        :type buff: :class:`bytearray`
        :param offset: Offset into the message
        :type offset: int
        """
        fmt = '[S [ihqY] ]'
        response = struct_helpers.unpack_from(fmt, buff, offset)
        self.topics = defaultdict(dict)
        for (topic, partitions) in response:
            for partition in partitions:
                self.topics[topic][partition[0]] = FetchPartitionResponse(
                    partition[2],
                    self._unpack_message_set(partition[3],
                                             partition_id=partition[0],
                                             broker_version=broker_version),
                    partition[1]
                )

    def _unpack_message_set(self, buff, partition_id=-1, broker_version='0.9.0'):
        """MessageSets can be nested. Get just the Messages out of it."""
        output = []
        message_set = MessageSet.decode(buff, partition_id=partition_id)
        for message in message_set.messages:
            if message.compression_type == CompressionType.NONE:
                output.append(message)
                continue
            elif message.compression_type == CompressionType.GZIP:
                decompressed = compression.decode_gzip(message.value)
                messages = self._unpack_message_set(decompressed,
                                                    partition_id=partition_id)
            elif message.compression_type == CompressionType.SNAPPY:
                decompressed = compression.decode_snappy(message.value)
                messages = self._unpack_message_set(decompressed,
                                                    partition_id=partition_id)
            elif message.compression_type == CompressionType.LZ4:
                if parse_version(broker_version) >= parse_version('0.10.0'):
                    decompressed = compression.decode_lz4(message.value)
                else:
                    decompressed = compression.decode_lz4_old_kafka(message.value)
                messages = self._unpack_message_set(decompressed,
                                                    partition_id=partition_id)
            if messages[-1].offset < message.offset:
                # With protocol 1, offsets from compressed messages start at 0
                assert messages[0].offset == 0
                delta = message.offset - len(messages) + 1
                for msg in messages:
                    msg.offset += delta
            output += messages
        return output


class FetchResponseV1(FetchResponse):
    API_VERSION = 1

    def __init__(self, buff, offset=0, broker_version='0.9.0'):
        """Deserialize into a new Response

        :param buff: Serialized message
        :type buff: :class:`bytearray`
        :param offset: Offset into the message
        :type offset: int
        """
        # TODO: Use throttle_time
        self.throttle_time = struct_helpers.unpack_from("i", buff, offset)[0]
        super(FetchResponseV1, self).__init__(buff, offset + 4,
                                              broker_version=broker_version)


class FetchResponseV2(FetchResponseV1):
    API_VERSION = 2



import logging

from .broker import Broker
from .simpleconsumer import SimpleConsumer
from .cluster import Cluster
from .partition import Partition
from .producer import Producer
from .topic import Topic
from .connection import SslConfig
from .client import KafkaClient
from .balancedconsumer import BalancedConsumer
from .managedbalancedconsumer import ManagedBalancedConsumer
from .membershipprotocol import RangeProtocol, RoundRobinProtocol

__version__ = '2.8.0'


__all__ = ["Broker", "SimpleConsumer", "Cluster", "Partition", "Producer",
           "Topic", "SslConfig", "KafkaClient", "BalancedConsumer",
           "ManagedBalancedConsumer", "RangeProtocol", "RoundRobinProtocol"]

logging.getLogger(__name__).addHandler(logging.NullHandler())



import itertools
import logging
from collections import namedtuple

from .protocol import ConsumerGroupProtocolMetadata

log = logging.getLogger(__name__)

GroupMembershipProtocol = namedtuple(
    'GroupMembershipProtocol', ['protocol_type',
                                'protocol_name',
                                'metadata',
                                'decide_partitions'])


def decide_partitions_range(participants, partitions, consumer_id):
    """Decide which partitions belong to this consumer_id.

    Uses the consumer rebalancing algorithm described here
    https://kafka.apache.org/documentation/#impl_consumerrebalance

    It is very important that the participants array is sorted,
    since this algorithm runs on each consumer and indexes into the same
    array. The same array index operation must return the same
    result on each consumer.

    :param participants: Sorted list of ids of all consumers in this
        consumer group.
    :type participants: Iterable of `bytes`
    :param partitions: List of all partitions on the topic being consumed
    :type partitions: Iterable of :class:`pykafka.partition.Partition`
    :param consumer_id: The ID of the consumer for which to generate a partition
        assignment.
    :type consumer_id: bytes
    """
    # Freeze and sort partitions so we always have the same results
    def p_to_str(p):
        return '-'.join([str(p.topic.name), str(p.leader.id), str(p.id)])

    all_parts = sorted(partitions.values(), key=p_to_str)

    # get start point, # of partitions, and remainder
    participants = sorted(participants)  # just make sure it's sorted.
    idx = participants.index(consumer_id)
    parts_per_consumer = len(all_parts) // len(participants)
    remainder_ppc = len(all_parts) % len(participants)

    start = parts_per_consumer * idx + min(idx, remainder_ppc)
    num_parts = parts_per_consumer + (0 if (idx + 1 > remainder_ppc) else 1)

    # assign partitions from i*N to (i+1)*N - 1 to consumer Ci
    new_partitions = itertools.islice(all_parts, start, start + num_parts)
    new_partitions = set(new_partitions)
    log.info('%s: Balancing %i participants for %i partitions. Owning %i partitions.',
             consumer_id, len(participants), len(all_parts),
             len(new_partitions))
    log.debug('My partitions: %s', [p_to_str(p) for p in new_partitions])
    return new_partitions


RangeProtocol = GroupMembershipProtocol(b"consumer",
                                        b"range",
                                        ConsumerGroupProtocolMetadata(),
                                        decide_partitions_range)


def decide_partitions_roundrobin(participants, partitions, consumer_id):
    """Decide which partitions belong to this consumer_id.

    Uses the "roundrobin" strategy described here
    https://kafka.apache.org/documentation/#oldconsumerconfigs

    :param participants: Sorted list of ids of all consumers in this
        consumer group.
    :type participants: Iterable of `bytes`
    :param partitions: List of all partitions on the topic being consumed
    :type partitions: Iterable of :class:`pykafka.partition.Partition`
    :param consumer_id: The ID of the consumer for which to generate a partition
        assignment.
    :type consumer_id: bytes
    """
    # Freeze and sort partitions so we always have the same results
    def p_to_str(p):
        return '-'.join([str(p.topic.name), str(p.leader.id), str(p.id)])

    partitions = sorted(partitions.values(), key=p_to_str)
    participants = sorted(participants)

    new_partitions = set()
    partitions_idx = participants_idx = 0
    for _ in range(len(partitions)):
        if participants[participants_idx] == consumer_id:
            new_partitions.add(partitions[partitions_idx])
        partitions_idx = (partitions_idx + 1) % len(partitions)
        participants_idx = (participants_idx + 1) % len(participants)

    log.info('%s: Balancing %i participants for %i partitions. Owning %i partitions.',
             consumer_id, len(participants), len(partitions),
             len(new_partitions))
    log.debug('My partitions: %s', [p_to_str(p) for p in new_partitions])
    return new_partitions


RoundRobinProtocol = GroupMembershipProtocol(b"consumer",
                                             b"roundrobin",
                                             ConsumerGroupProtocolMetadata(),
                                             decide_partitions_roundrobin)



#!/usr/bin/env python
__license__ = """
Copyright 2015 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re
import sys
import os
import platform

from distutils.command.build_ext import build_ext
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from setuptools.extension import Extension

if sys.version_info < (2, 7):
    raise Exception('pykafka requires Python 2.7 or higher.')

python_implementation = platform.python_implementation()
is_cpython = python_implementation == 'CPython'

# Get version without importing, which avoids dependency issues
def get_version():
    with open('pykafka/__init__.py') as version_file:
        return re.search(r"""__version__\s+=\s+(['"])(?P<version>.+?)\1""",
                         version_file.read()).group('version')

install_requires = [
    'six>=1.5',
    'kazoo==2.5.0',
    'tabulate'
]

extra_gevent_requires = [
    'gevent==1.3'
]

extra_requires = [
    'lz4==0.10.1',
    'lz4tools==1.3.1.2',
    'xxhash==1.0.1'
] + extra_gevent_requires

lint_requires = [
    'pep8',
    'pyflakes'
]


def read_lines(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.readlines()

tests_require = [
    x.strip() for x in read_lines('test-requirements.txt') if not x.startswith('-')
]

dependency_links = []
setup_requires = []


class ve_build_ext(build_ext):
    """This class allows C extension building to fail.

    If the name seems horribly random: we've kept it for historic reasons, see
    http://nedbatchelder.com/blog/201212/skipping_c_extensions.html
    This particular evolution was honestly stolen from the sqlalchemy project.
    """
    class BuildFailed(Exception):
        def __init__(self):
            self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise self.BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except (CCompilerError, DistutilsExecError, DistutilsPlatformError):
            raise self.BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise self.BuildFailed()
            raise


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

if 'nosetests' in sys.argv[1:]:
    setup_requires.append('nose')

with open('README.rst') as f:
    readme = f.read()

def run_setup(with_rdkafka=True):
    """This just calls setuptools.setup() but makes ext_modules optional"""
    ext_modules = []
    if with_rdkafka:
        ext_modules.append(Extension(
            'pykafka.rdkafka._rd_kafka',
            libraries=['rdkafka'],
            sources=['pykafka/rdkafka/_rd_kafkamodule.c']))

    setup(
        name='pykafka',
        version=get_version(),
        author='Keith Bourgoin and Emmett Butler',
        author_email='pykafka-user@googlegroups.com',
        url='https://github.com/Parsely/pykafka',
        description='Full-Featured Pure-Python Kafka Client',
        long_description=readme,
        keywords='apache kafka client driver',
        license='Apache License 2.0',
        packages=find_packages(),
        entry_points={
            'console_scripts': [
                'kafka-tools = pykafka.cli.kafka_tools:main',
            ]
        },
        install_requires=install_requires,
        tests_require=tests_require,
        setup_requires=setup_requires,
        extras_require={
            'test': tests_require,
            'all': install_requires + tests_require + extra_requires,
            'docs': ['sphinx'] + tests_require,
            'lint': lint_requires,
            'gevent': extra_gevent_requires
        },
        cmdclass={'test': PyTest, 'build_ext': ve_build_ext},
        ext_modules=ext_modules,
        dependency_links=dependency_links,
        zip_safe=False,
        test_suite='nose.collector',
        include_package_data=True,
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Programming Language :: Python",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Topic :: Database",
            "Topic :: Database :: Front-Ends",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ]
    )

try:
    if not is_cpython:
        print("librdkafka is not supported under %s" % python_implementation)
        print(15 * "-")
        print("INFO: Failed to build rdkafka extension:")
        print("INFO: will now attempt setup without extension.")
        print(15 * "-")
        run_setup(with_rdkafka=False)
    else:
        run_setup()
except ve_build_ext.BuildFailed as exc:
    print(15 * "-")
    print("INFO: Failed to build rdkafka extension:")
    print(exc.cause)
    print("INFO: will now attempt setup without extension.")
    print(15 * "-")
    run_setup(with_rdkafka=False)



# - coding: utf-8 -
from .admin import (ListGroupsRequest, ListGroupsResponse,
                    DescribeGroupsRequest, DescribeGroupsResponse,
                    CreateTopicsRequest, CreateTopicsResponse, DeleteTopicsRequest,
                    DeleteTopicsResponse, ApiVersionsRequest, ApiVersionsResponse,
                    CreateTopicRequest)
from .base import API_VERSIONS_080, API_VERSIONS_090
from .fetch import (PartitionFetchRequest, FetchRequest, FetchPartitionResponse,
                    FetchResponse, FetchResponseV1, FetchResponseV2)
from .group_membership import (JoinGroupRequest, JoinGroupResponse, SyncGroupRequest,
                               SyncGroupResponse, HeartbeatRequest, HeartbeatResponse,
                               LeaveGroupRequest, LeaveGroupResponse,
                               ConsumerGroupProtocolMetadata, MemberAssignment)
from .message import Message, MessageSet
from .metadata import (MetadataRequest, MetadataResponse, MetadataRequestV1,
                       MetadataResponseV1, MetadataRequestV2, MetadataResponseV2,
                       MetadataRequestV3, MetadataResponseV3, MetadataResponseV4,
                       MetadataRequestV4, MetadataRequestV5, MetadataResponseV5)
from .offset import (ListOffsetRequest, ListOffsetResponse, PartitionOffsetRequest,
                     ListOffsetRequestV1, ListOffsetResponseV1)
from .offset_commit import (GroupCoordinatorRequest, GroupCoordinatorResponse,
                            PartitionOffsetCommitRequest, OffsetCommitRequest,
                            OffsetCommitPartitionResponse, OffsetCommitResponse,
                            PartitionOffsetFetchRequest, OffsetFetchRequest,
                            OffsetFetchPartitionResponse, OffsetFetchResponse,
                            OffsetFetchRequestV1, OffsetFetchResponseV1,
                            OffsetFetchRequestV2, OffsetFetchResponseV2)
from .produce import ProduceRequest, ProduceResponse, ProducePartitionResponse

"""
Author: Keith Bourgoin, Emmett Butler

Protocol implementation for Kafka>=0.8.2

The implementation has been done with an attempt to minimize memory
allocations in order to improve performance. With the exception of
compressed messages, we can calculate the size of the entire message
to send and do only a single memory allocation.

For Reference:

https://cwiki.apache.org/confluence/display/KAFKA/A+Guide+To+The+Kafka+Protocol

Each message is encoded as either a Request or Response:

RequestOrResponse => Size (RequestMessage | ResponseMessage)
  Size => int32

RequestMessage => ApiKey ApiVersion CorrelationId ClientId RequestMessage
  ApiKey => int16
  ApiVersion => int16
  CorrelationId => int32
  ClientId => string
  RequestMessage => MetadataRequest | ProduceRequest | FetchRequest | OffsetRequest | OffsetCommitRequest | OffsetFetchRequest

Response => CorrelationId ResponseMessage
  CorrelationId => int32
  ResponseMessage => MetadataResponse | ProduceResponse | FetchResponse | OffsetResponse | OffsetCommitResponse | OffsetFetchResponse
"""
__all__ = ["MetadataRequest", "MetadataResponse", "ProduceRequest", "ProduceResponse",
           "PartitionFetchRequest", "FetchRequest", "FetchPartitionResponse",
           "FetchResponse", "ListOffsetRequest", "ListOffsetResponse",
           "GroupCoordinatorRequest", "GroupCoordinatorResponse",
           "PartitionOffsetCommitRequest", "OffsetCommitRequest",
           "OffsetCommitPartitionResponse", "OffsetCommitResponse",
           "PartitionOffsetFetchRequest", "OffsetFetchRequest",
           "OffsetFetchPartitionResponse", "OffsetFetchResponse",
           "JoinGroupRequest", "JoinGroupResponse", "SyncGroupRequest",
           "SyncGroupResponse", "HeartbeatRequest", "HeartbeatResponse",
           "LeaveGroupRequest", "LeaveGroupResponse", "ListGroupsRequest",
           "ListGroupsResponse", "DescribeGroupsRequest", "DescribeGroupsResponse",
           "Message", "MessageSet", "ApiVersionsRequest", "ApiVersionsResponse",
           "CreateTopicsRequest", "CreateTopicsResponse", "DeleteTopicsRequest",
           "DeleteTopicsResponse", "PartitionOffsetRequest", "API_VERSIONS_080",
           "API_VERSIONS_090", "ConsumerGroupProtocolMetadata", "MemberAssignment",
           "FetchResponseV1", "FetchResponseV2", "MetadataResponseV1",
           "MetadataRequestV1", "CreateTopicRequest", "ProducePartitionResponse",
           "ListOffsetRequestV1", "ListOffsetResponseV1", "OffsetFetchRequestV1",
           "OffsetFetchResponseV1", "OffsetFetchRequestV2", "OffsetFetchResponseV2",
           "MetadataRequestV2", "MetadataResponseV2", "MetadataRequestV3",
           "MetadataResponseV3", "MetadataRequestV4", "MetadataResponseV4",
           "MetadataRequestV5", "MetadataResponseV5"]



"""
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import print_function

import argparse
import errno
import logging
import os
import signal
import shutil
import socket
import subprocess
import sys
import tempfile
import time

from testinstances import utils
from testinstances.exceptions import ProcessNotStartingError
from testinstances.managed_instance import ManagedInstance
from pykafka.utils.compat import range, get_bytes, get_string


log = logging.getLogger(__name__)

_kafka_properties = """
# Configurable settings
broker.id={broker_id}
{port_config}
zookeeper.connect={zk_connstr}
log.dirs={data_dir}

# non-configurable settings
num.network.threads=3
num.io.threads=8
socket.send.buffer.bytes=102400
socket.receive.buffer.bytes=102400
socket.request.max.bytes=104857600
num.partitions=1
num.recovery.threads.per.data.dir=1
log.segment.bytes=1073741824
log.cleaner.enable=false
zookeeper.connection.timeout.ms=6000
delete.topic.enable=true
"""

_kafka_ssl_properties = """
listeners=PLAINTEXT://localhost:{port},SSL://localhost:{ssl_port}
ssl.keystore.location={keystore_path}
ssl.keystore.password={store_pass}
ssl.key.password={store_pass}
ssl.truststore.location={truststore_path}
ssl.truststore.password={store_pass}
"""

_zookeeper_properties = """
dataDir={zk_dir}
clientPort={zk_port}
maxClientCnxns=0
"""

class KafkaConnection(object):
    """Connection to a Kafka cluster.

    Provides handy access to the shell scripts Kafka is bundled with.
    """

    def __init__(self, bin_dir, brokers, zookeeper, brokers_ssl=None):
        """Create a connection to the cluster.

        :param bin_dir:   Location of downloaded kafka bin
        :param brokers:   Comma-separated list of brokers
        :param zookeeper: Connection straing for ZK
        :param brokers_ssl: Comma-separated list of hosts with ssl-ports
        """
        self._bin_dir = bin_dir
        self.brokers = brokers
        self.zookeeper = zookeeper

        self.brokers_ssl = brokers_ssl
        self.certs = CertManager(bin_dir) if brokers_ssl is not None else None

    def _run_topics_sh(self, args):
        """Run kafka-topics.sh with the provided list of arguments."""
        binfile = os.path.join(self._bin_dir, 'bin/kafka-topics.sh')
        cmd = [binfile, '--zookeeper', self.zookeeper] + args
        cmd = [get_string(c) for c in cmd]  # execv needs only strings
        log.debug('running: %s', ' '.join(cmd))
        return subprocess.check_output(cmd)

    def create_topic(self, topic_name, num_partitions, replication_factor):
        """Use kafka-topics.sh to create a topic."""
        log.info('Creating topic %s', topic_name)
        self._run_topics_sh(['--create',
                             '--topic', topic_name,
                             '--partitions', num_partitions,
                             '--replication-factor', replication_factor])
        time.sleep(2)

    def delete_topic(self, topic_name):
        self._run_topics_sh(['--delete',
                             '--topic', topic_name])

    def flush(self):
        """Delete all topics."""
        for topic in self.list_topics():
            if not topic.startswith(b'__'):  # leave internal topics alone
                self.delete_topic(topic)

    def list_topics(self):
        """Use kafka-topics.sh to get topic information."""
        res = self._run_topics_sh(['--list'])
        return res.strip().split(b'\n')

    def produce_messages(self, topic_name, messages, batch_size=200):
        """Produce some messages to a topic."""
        binfile = os.path.join(self._bin_dir, 'bin/kafka-console-producer.sh')
        cmd = [binfile,
               '--broker-list', self.brokers,
               '--topic', topic_name,
               '--batch-size', batch_size]
        cmd = [get_string(c) for c in cmd]  # execv needs only strings
        log.debug('running: %s', ' '.join(cmd))
        proc = subprocess.Popen(cmd, stdin=subprocess.PIPE)
        proc.communicate(input=get_bytes('\n'.join(messages)))
        if proc.poll() is None:
            proc.kill()


class KafkaInstance(ManagedInstance):
    """A managed kafka instance for testing"""

    def __init__(self,
                 num_instances=1,
                 kafka_version='0.8.2.1',
                 scala_version='2.11',
                 bin_dir='/tmp/kafka-bin',
                 name='kafka',
                 use_gevent=False):
        """Start kafkainstace with given settings"""
        self._num_instances = num_instances
        self._kafka_version = kafka_version
        self._scala_version = scala_version
        self._bin_dir = bin_dir
        self._processes = []
        self._broker_procs = []
        self._brokers_started = 0  # incremented by _start_broker
        self.zookeeper = None
        self.brokers = None
        self.brokers_ssl = None
        self.certs = self._gen_ssl_certs()
        # TODO: Need a better name so multiple can run at once.
        #       other ManagedInstances use things like 'name-port'
        ManagedInstance.__init__(self, name, use_gevent=use_gevent)
        self.connection = KafkaConnection(
            bin_dir, self.brokers, self.zookeeper, self.brokers_ssl)

    def _init_dirs(self):
        """Set up directories in the temp folder."""
        self._conf_dir = os.path.join(self._root_dir, 'conf')
        self._data_dir = os.path.join(self._root_dir, 'data')
        self._log_dir = os.path.join(self._root_dir, 'logs')
        self._zk_dir = os.path.join(self._root_dir, 'zk')
        os.makedirs(self._conf_dir)
        os.makedirs(self._data_dir)
        os.makedirs(self._log_dir)
        os.makedirs(self._zk_dir)
        # Needs to exist so that testinstances doesn't break
        open(self.logfile, 'w').close()

    def _download_kafka(self):
        """Make sure the Kafka code has been downloaded to the right dir."""
        binfile = os.path.join(self._bin_dir,
                               'bin/kafka-server-start.sh')
        if os.path.exists(binfile):
            return  # already there

        # Make the download dir
        try:
            os.makedirs(self._bin_dir)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

        # Download and extract
        log.info('Downloading Kafka.')
        curr_dir = os.getcwd()
        os.chdir(self._bin_dir)
        url_fmt = 'https://archive.apache.org/dist/kafka/{kafka_version}/kafka_{scala_version}-{kafka_version}.tgz'
        url = url_fmt.format(
            scala_version=self._scala_version,
            kafka_version=self._kafka_version
        )
        p1 = subprocess.Popen(['curl', '-vs', url], stdout=subprocess.PIPE)
        p2 = subprocess.Popen(['tar', 'xvz', '-C', self._bin_dir,
                               '--strip-components', '1'],
                              stdin=p1.stdout, stdout=subprocess.PIPE)
        p1.stdout.close()
        output, err = p2.communicate()
        os.chdir(curr_dir)

        log.info('Downloaded Kafka to %s', self._bin_dir)

    def _is_port_free(self, port):
        """Check to see if a port is open"""
        try:
            s = socket.create_connection(('localhost', port))
            s.close()
            return False
        except IOError as err:
            return err.errno == errno.ECONNREFUSED

    def _port_generator(self, start):
        """Generate open ports, starting with `start`."""
        port = start
        while True:
            if self._is_port_free(port):
                yield port
            port += 1

    def _add_broker(self, broker_port):
        new_broker = "localhost:{}".format(broker_port)
        brokers = self.brokers.split(",") if self.brokers else []
        brokers.append(new_broker)
        self.brokers = ",".join(brokers)

    def _add_ssl_broker(self, ssl_broker_port):
        if ssl_broker_port:
            new_broker_ssl = "localhost:{}".format(ssl_broker_port)
            brokers_ssl = self.brokers_ssl.split(",") if self.brokers_ssl else []
            brokers_ssl.append(new_broker_ssl)
            self.brokers_ssl = ",".join(brokers_ssl)

    def _gen_ssl_certs(self):
        """Attempt generating ssl certificates for testing

        :returns: :class:`CertManager` or None upon failure
        """
        if self._kafka_version >= "0.9":  # no SSL support in earlier versions
            try:
                return CertManager(self._bin_dir)
            except:  # eg. because openssl or other tools not installed
                log.exception("Couldn't generate ssl certs:")

    def _start_process(self):
        """Start the instance processes"""
        self._init_dirs()
        self._download_kafka()

        # Start all relevant processes and save which ports they use
        zk_port = self._start_zookeeper()
        self.zookeeper = 'localhost:{port}'.format(port=zk_port)

        broker_ports, broker_ssl_ports = self._start_brokers()

        # Process is started when the port isn't free anymore
        all_ports = [zk_port] + broker_ports
        for i in range(10):
            if all(not self._is_port_free(port) for port in all_ports):
                log.info('Kafka cluster started.')
                return  # hooray! success
            log.info('Waiting for cluster to start....')
            time.sleep(6)  # Waits 60s total

        # If it got this far, it's an error
        raise ProcessNotStartingError('Unable to start Kafka cluster.')

    def _start_log_watcher(self):
        """Overridden because we have multiple files to watch."""
        for logfile in os.listdir(self._log_dir):
            logfile = open(os.path.join(self._log_dir, logfile), 'r')
            if self.use_gevent:
                import gevent
                gevent.spawn(self._watcher_gevent, logfile)
            else:
                import threading
                watch_thread = threading.Thread(
                    target=self._watcher_threading, args=[logfile]
                )
                watch_thread.daemon = True
                watch_thread.start()

    def _start_broker_proc(self, port, ssl_port=None):
        """Start a broker proc and maintain handlers

        Returns a proc handler for the new broker.
        """
        # make port config for new broker
        if self.certs is not None:
            port_config = _kafka_ssl_properties.format(
                port=port,
                ssl_port=ssl_port,
                keystore_path=self.certs.keystore,
                truststore_path=self.certs.truststore,
                store_pass=self.certs.broker_pass)
        else:
            port_config = "port={}".format(port)

        self._brokers_started += 1
        i = self._brokers_started

        # write conf file for the new broker
        conf = os.path.join(self._conf_dir,
                            'kafka_{instance}.properties'.format(instance=i))
        with open(conf, 'w') as f:
            f.write(_kafka_properties.format(
                broker_id=i,
                port_config=port_config,
                zk_connstr=self.zookeeper,
                data_dir=self._data_dir + '_{instance}'.format(instance=i),
            ))

        # start process and append to self._broker_procs
        binfile = os.path.join(self._bin_dir, 'bin/kafka-server-start.sh')
        logfile = os.path.join(self._log_dir, 'kafka_{instance}.log'.format(instance=i))
        new_proc = (utils.Popen(
            args=[binfile, conf],
            stderr=utils.STDOUT,
            stdout=open(logfile, 'w'),
            use_gevent=self.use_gevent
        ))
        self._broker_procs.append(new_proc)

        # add localhost:port to internal list of (ssl)brokers
        self._add_broker(port)
        self._add_ssl_broker(ssl_port)

        return new_proc


    def _start_brokers(self):
        """Start all brokers and return used ports."""
        ports = self._port_generator(9092)
        used_ports = []
        used_ssl_ports = []
        ssl_port = None
        for i in range(self._num_instances):
            port = next(ports)
            used_ports.append(port)
            log.info('Starting Kafka on port %i.', port)

            if self.certs is not None:
                ssl_port = next(ports)
                used_ssl_ports.append(ssl_port)  # to return at end
            self._start_broker_proc(port, ssl_port)

        return used_ports, used_ssl_ports

    def _start_zookeeper(self):
        port = next(self._port_generator(2181))
        log.info('Starting zookeeper on port %i.', port)

        conf = os.path.join(self._conf_dir, 'zk.properties')
        with open(conf, 'w') as f:
            f.write(_zookeeper_properties.format(zk_dir=self._zk_dir,
                                                 zk_port=port))

        binfile = os.path.join(self._bin_dir, 'bin/zookeeper-server-start.sh')
        logfile = os.path.join(self._log_dir, 'zk.log')
        self._zk_proc = utils.Popen(
            args=[binfile, conf],
            stderr=utils.STDOUT,
            stdout=open(logfile, 'w'),
            use_gevent=self.use_gevent
        )
        return port

    def terminate(self):
        """Override because we have many processes."""
        # Kill brokers before zookeeper
        log.info('Stopping Kafka brokers.')
        for broker in self._broker_procs:
            broker.kill()
        log.info('Stopping Zookeeper.')
        self._zk_proc.kill()

        shutil.rmtree(self._root_dir)
        self._processes = []
        super(KafkaInstance, self).terminate()
        log.info('Cluster terminated and data cleaned up.')

    def flush(self):
        """Flush all data in the cluster"""
        raise NotImplementedError()

    def create_topic(self, topic_name, num_partitions, replication_factor):
        """Use kafka-topics.sh to create a topic."""
        return self.connection.create_topic(topic_name, num_partitions,
                                            replication_factor)

    def delete_topic(self, topic_name):
        return self.connection.delete_topic(topic_name)

    def list_topics(self):
        """Use kafka-topics.sh to get topic information."""
        return self.connection.list_topics()

    def produce_messages(self, topic_name, messages):
        """Produce some messages to a topic."""
        return self.connection.produce_messages(topic_name, messages)


class CertManager(object):
    """Helper that generates key pairs/stores for brokers and clients

    NB: intended only for test suite purposes.  For instance, currently the
    same cert is shared between all brokers, and another between all clients,
    even between different instances of CertManager.  And for another, this
    class freely discloses passwords for its keys: they're purely here to
    test if the configuration with passwords works
    """

    def __init__(self, root_dir):
        # It would be more elegant to use a tempdir here, but that would make
        # it difficult for KafkaConnection to find the CA cert for an already
        # running cluster, and so we just use a fixed location:
        self._dir = os.path.join(root_dir, "test-certs")

        self.validity_days = '10000'

        self.root_cert = os.path.join(self._dir, "root.cert")
        self.root_key = os.path.join(self._dir, "root.key")
        self.root_pass = "NfXI63pmNqbHF4tq"

        self.truststore = os.path.join(self._dir, "broker.truststore.jks")
        self.keystore = os.path.join(self._dir, "broker.keystore.jks")
        self.broker_pass = "JJQqPLUV1ZIFD6ny"

        self.client_cert = os.path.join(self._dir, "client.cert")
        self.client_key = os.path.join(self._dir, "client.key")
        self.client_pass = "dWJTxlilbJcH1Wgn"

        if not os.path.exists(self.root_cert):
            try:
                os.makedirs(self._dir)
            except OSError:
                pass
            self._gen_root_cert()
            self._gen_broker_truststore()
            self._gen_broker_keystore()
            self._gen_client_cert()

    def _gen_root_cert(self):
        """Generate root certificate"""
        cmd = ['openssl', 'req',
               '-x509',
               '-newkey', 'rsa:2048',
               '-subj', '/CN=ca.example.com',
               '-days', self.validity_days,
               '-out', self.root_cert,
               '-keyout', self.root_key,
               '-passout', 'env:ROOT_PASS']
        return subprocess.check_call(cmd, env=dict(ROOT_PASS=self.root_pass))

    def _gen_broker_truststore(self):
        """Generate truststore that trusts self.root_cert"""
        cmd = ['keytool', '-importcert', '-alias', 'root',
                                         '-file', self.root_cert,
                                         '-keystore', self.truststore,
                                         '-storepass:env', 'BROKER_PASS',
                                         '-noprompt']
        env = dict(BROKER_PASS=self.broker_pass)
        return subprocess.check_call(cmd, env=env)

    def _gen_broker_keystore(self):
        """Generate broker keypair, signed by self.root_cert"""
        env = dict(BROKER_PASS=self.broker_pass, ROOT_PASS=self.root_pass)
        csr_path = os.path.join(self._dir, "broker.csr")
        cert_path = os.path.join(self._dir, "broker.cert")

        cmd = ['keytool', '-genkeypair', '-alias', 'broker',
                                         '-dname', 'CN=localhost',
                                         '-validity', self.validity_days,
                                         '-keystore', self.keystore,
                                         '-storepass:env', 'BROKER_PASS',
                                         '-keypass:env', 'BROKER_PASS',
                                         '-noprompt']
        subprocess.check_call(cmd, env=env)
        cmd = ['keytool', '-certreq', '-alias', 'broker',
                                      '-file', csr_path,
                                      '-keystore', self.keystore,
                                      '-storepass:env', 'BROKER_PASS']
        subprocess.check_call(cmd, env=env)
        cmd = ['openssl', 'x509', '-req', '-in', csr_path,
                                          '-out', cert_path,
                                          '-days', self.validity_days,
                                          '-CA', self.root_cert,
                                          '-CAkey', self.root_key,
                                          '-CAcreateserial',
                                          '-passin', 'env:ROOT_PASS']
        subprocess.check_call(cmd, env=env)
        cmd = ['keytool', '-importcert', '-alias', 'root',
                                         '-file', self.root_cert,
                                         '-keystore', self.keystore,
                                         '-storepass:env', 'BROKER_PASS',
                                         '-noprompt']
        subprocess.check_call(cmd, env=env)
        cmd = ['keytool', '-importcert', '-alias', 'broker',
                                         '-file', cert_path,
                                         '-keystore', self.keystore,
                                         '-storepass:env', 'BROKER_PASS']
        subprocess.check_call(cmd, env=env)

        os.unlink(csr_path)
        os.unlink(cert_path)

    def _gen_client_cert(self):
        """Generate client keypair and sign it"""
        env = dict(CLIENT_PASS=self.client_pass, ROOT_PASS=self.root_pass)
        csr_path = os.path.join(self._dir, "client.csr")

        cmd = ['openssl', 'req', '-newkey', 'rsa:2048',
                                 '-subj', '/CN=localhost',
                                 '-days', self.validity_days,
                                 '-out', csr_path,
                                 '-keyout', self.client_key,
                                 '-passout', 'env:CLIENT_PASS']
        subprocess.check_call(cmd, env=env)
        cmd = ['openssl', 'x509', '-req', '-in', csr_path,
                                          '-out', self.client_cert,
                                          '-days', self.validity_days,
                                          '-CA', self.root_cert,
                                          '-CAkey', self.root_key,
                                          '-passin', 'env:ROOT_PASS']
        subprocess.check_call(cmd, env=env)

        os.unlink(csr_path)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog='kafka_instance',
        usage='foo bars',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('num_brokers', type=int, default=3,
                        help='Numbers of brokers in the cluster.')
    parser.add_argument('--download-dir', type=str, default='/tmp/kafka-bin',
                        help='Download destination for Kafka')
    parser.add_argument('--kafka-version', type=str, default='0.8.2.1',
                        help='Kafka version to download')
    parser.add_argument('--scala-version', type=str, default='2.11',
                        help='Scala version for kafka build')
    parser.add_argument('--export-hosts', type=str,
                        help='Write host strings to given file path')
    args = parser.parse_args()

    _exiting = False
    def _catch_sigint(signum, frame):
        global _exiting
        _exiting = True
        print('SIGINT received.')
    signal.signal(signal.SIGINT, _catch_sigint)

    cluster = KafkaInstance(num_instances=args.num_brokers,
                            kafka_version=args.kafka_version,
                            scala_version=args.scala_version,
                            bin_dir=args.download_dir)
    print('Cluster started.')
    print('Brokers: {brokers}'.format(brokers=cluster.brokers))
    print('Zookeeper: {zk}'.format(zk=cluster.zookeeper))
    print('Waiting for SIGINT to exit.')

    if args.export_hosts is not None:
        with open(args.export_hosts, 'w') as f:
            f.write('BROKERS={}\n'.format(cluster.brokers))
            if cluster.brokers_ssl:
                f.write('BROKERS_SSL={}\n'.format(cluster.brokers_ssl))
            f.write('ZOOKEEPER={}\n'.format(cluster.zookeeper))

    while True:
        if _exiting:
            print('Exiting.')
            sys.exit(0)
        time.sleep(1)
